//////////////////////////////////////////////////////////////////////
// driver_configurator.cpp: implementation of the driver_configurator class.
//
// Author	:	David Rokhvarg	<davidr@sangoma.com>
//////////////////////////////////////////////////////////////////////

#include "driver_configurator.h"

#define DBG_CFG		if(0)printf("DRVCFG:");if(0)printf
#define _DBG_CFG	if(0)printf

#define INFO_CFG	if(0)printf("DRVCFG:");if(0)printf
#define _INFO_CFG	if(0)printf

#define ERR_CFG		if(1)printf("DRVCFG:");if(1)printf
#define _ERR_CFG	if(1)printf

#if !defined(__FUNCTION__)
#define	__FUNCTION__ "function"
#endif
//////////////////////////////////////////////////////////////////////
//common functions
//////////////////////////////////////////////////////////////////////

//these are recommendations only:
#define MIN_RECOMMENDED_BITSTREAM_MTU	256
#define MIN_RECOMMENDED_HDLC_MTU		2048

static int get_number_of_channels(unsigned int chan_bit_map)
{
	int chan_counter = 0, i;

	for(i = 0; i < sizeof(unsigned int)*8; i++){
		if(chan_bit_map & (1 << i)){
			chan_counter++;
		}
	}
	return chan_counter;
}

static int get_recommended_bitstream_mtu(int num_of_channels)
{
	int recommended_mtu = num_of_channels * 32;//<--- will work well in most cases

	//
	//Smaller mtu will cause more interrupts.
	//Recommend at least MIN_RECOMMENDED_BITSTREAM_MTU, user will set to smaller, if timing is important.
	//
	while(recommended_mtu < MIN_RECOMMENDED_BITSTREAM_MTU){
		recommended_mtu = recommended_mtu * 2;
	}

	return recommended_mtu;
}

static int check_mtu(int mtu, int hdlc_streaming, unsigned int channel_bitmap)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "check_mtu";
#endif
	DBG_CFG("%s()\n", __FUNCTION__);

	DBG_CFG("channel_bitmap: 0x%X\n", channel_bitmap);

	if(mtu % 4){
		ERR_CFG("Invalid MTU (%d). Must be divizible by 4.\n", mtu);
		return 1;
	}

	//if BitStream, check MTU is also divisible by the number of channels (timeslots) in the group
	if(hdlc_streaming == WANOPT_NO){

		int num_of_channels = get_number_of_channels(channel_bitmap);
	
		DBG_CFG("num_of_channels in bitmap: %d\n", num_of_channels);

		if(mtu % num_of_channels){
ERR_CFG(
"Invalid MTU. For 'BitStream' Must be divisible by 4 AND by the number of\n\
\tchannels in the group (%d). Recommended MTU: %d.\n",
	num_of_channels, get_recommended_bitstream_mtu(num_of_channels));
			return 2;
		}
	}//if()

	return 0;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

driver_configurator::driver_configurator()
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "driver_configurator";
#endif
	DBG_CFG("%s()\n", __FUNCTION__);
	wp_handle = INVALID_HANDLE_VALUE;
}

driver_configurator::~driver_configurator()
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "~driver_configurator";
#endif
	DBG_CFG("%s()\n", __FUNCTION__);

	if(wp_handle != INVALID_HANDLE_VALUE){
		CloseHandle(wp_handle);
		wp_handle = INVALID_HANDLE_VALUE;
	}
}

int driver_configurator::get_configration(port_cfg_t *port_cfg)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "get_configration";
#endif
	DBG_CFG("%s()\n", __FUNCTION__);

	if(DoGetDriverConfigurationCommand(port_cfg)){
		return 1;
	}

	return port_cfg->operation_status;
}

//function to switch between T1 and E1
int driver_configurator::set_t1_e1_configuration(sdla_fe_cfg_t *sdla_fe_cfg, buffer_settings_t *buffer_settings)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "set_t1_e1_configuration";
#endif
	wandev_conf_t		*wandev_conf = &port_cfg.wandev_conf;
	wan_xilinx_conf_t	*wan_xilinx_conf = &wandev_conf->u.aft;

	DBG_CFG("%s()\n", __FUNCTION__);

	memset(&port_cfg, 0x00, sizeof(port_cfg_t));

	//copy T1/E1 configuration into the Driver configuration structure.
	memcpy(&wandev_conf->fe_cfg, sdla_fe_cfg, sizeof(sdla_fe_cfg_t));

	port_cfg.buffer_settings.buffer_multiplier_factor = buffer_settings->buffer_multiplier_factor;
	port_cfg.buffer_settings.number_of_buffers_per_api_interface = buffer_settings->number_of_buffers_per_api_interface;

	wandev_conf->fe_cfg.cfg.te_cfg.active_ch = 0xFFFFFFFF;//a constant
	port_cfg.num_of_ifs = 1;//1 Group
	strcpy(port_cfg.if_cfg[0].usedby, 
		SDLA_DECODE_USEDBY_FIELD(TDM_VOICE_API));//TDM_VOICE_API - the same as API, but 'BitSwap' is enabled.
		//SDLA_DECODE_USEDBY_FIELD(API_LEGACY));//API_LEGACY
	port_cfg.if_cfg[0].hdlc_streaming = WANOPT_NO;//BitStream
		
	if(sdla_fe_cfg->media == WAN_MEDIA_T1){
		//T1, 1 Group of 24 Timeslots.
		//	1-24 - in API mode, BitStream
		port_cfg.if_cfg[0].active_ch = 0xFFFFFF;
		port_cfg.if_cfg[0].mtu = 8184;//Maximum MTU/MRU, multiple of 24 timeslots
		FE_TDMV_LAW(sdla_fe_cfg) = WAN_TDMV_MULAW;

	}else if(sdla_fe_cfg->media == WAN_MEDIA_E1){
		//E1, 1 Group of 31 Timeslots.
		//	1-31 - in API mode, BitStream
		port_cfg.if_cfg[0].active_ch = 0xFFFFFFFF;
		if(FE_FRAME(sdla_fe_cfg) == WAN_FR_UNFRAMED){
			port_cfg.if_cfg[0].mtu = 8160;//Maximum MTU/MRU, multiple of 32 timeslots
		}else{
			port_cfg.if_cfg[0].mtu = 7936;//Maximum MTU/MRU, multiple of 31 timeslots
		}
		FE_TDMV_LAW(sdla_fe_cfg) = WAN_TDMV_ALAW;
		//FE_LBO(sdla_fe_cfg) = WAN_T1_LBO_0_DB;//important to set to a valid value!!
	}else{
		ERR_CFG("%s(): Error: invalid Media Type %d!\n", __FUNCTION__, sdla_fe_cfg->media);
		return 1;
	}

	//check correctness of the 'port_cfg' structure before trying to use it.
	if(check_port_cfg_structure(&port_cfg)){
		return 2;
	}
	
	print_port_cfg_structure(&port_cfg);

	return set_configration(&port_cfg);
}

int driver_configurator::set_configration(port_cfg_t *port_cfg)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "set_configration";
#endif
	DBG_CFG("%s()\n", __FUNCTION__);

	if(stop_or_start_port(STOP_PORT)){
		//failed to stop port for re-configuration
		return 1;
	}

	if(DoSetDriverConfigurationCommand(port_cfg)){
		//configration failed
		return 1;
	}

	INFO_CFG("%s(): return code: %s (%d)\n", __FUNCTION__,
		SDLA_DECODE_SANG_STATUS(port_cfg->operation_status), port_cfg->operation_status);
	switch(port_cfg->operation_status)
	{
	case SANG_STATUS_DEVICE_BUSY:
		INFO_CFG("Error: open handles exist for '%s_IF\?\?' interfaces!\n",	wanpipe_name_str);
		return port_cfg->operation_status;
	case SANG_STATUS_SUCCESS:
		//OK
		break;
	default:
		//error
		return 1;
	}

	if(stop_or_start_port(START_PORT_VOLATILE_CONFIG)){
		//failed to start port after re-configuration
		return 1;
	}

	return 0;
}


int driver_configurator::init(unsigned int wanpipe_number, unsigned int port_number)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "init";
#endif
	DBG_CFG("%s():\n", __FUNCTION__);

	wp_number = wanpipe_number + port_number + 1;
	_snprintf(wanpipe_name_str, WAN_DRVNAME_SZ*2, "\\\\.\\WANPIPE%d", 
		wp_number);
	DBG_CFG("%s\n", wanpipe_name_str);

	wp_handle = open_api_device(wanpipe_name_str);
	if(wp_handle == INVALID_HANDLE_VALUE){
//		ERR_CFG("Error: failed to open %s!!\n", wanpipe_name_str);
		return 1;
	}
	return 0;
}

HANDLE driver_configurator::open_api_device(char *api_device_name)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "open_api_device";
#endif
//	DBG_CFG("%s():%s\n", __FUNCTION__, api_device_name);

	return CreateFile(	api_device_name, 
						GENERIC_READ | GENERIC_WRITE, 
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						(LPSECURITY_ATTRIBUTES)NULL, 
						OPEN_EXISTING,
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH,
						(HANDLE)NULL
						);
}

int driver_configurator::DoPortConfigurationCommand(port_cfg_t *port_cfg)
{
	DWORD ln;
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "DoPortConfigurationCommand";
#endif

	if(DeviceIoControl(
			wp_handle,
			IoctlPortConfigurationCommand,
			(LPVOID)port_cfg,
			sizeof(port_cfg_t),
			(LPVOID)port_cfg,
			sizeof(port_cfg_t),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		ERR_CFG("%s():Error: IoctlSetDriverConfiguration failed!!\n", __FUNCTION__);
		return 1;
	}else{
		return 0;
	}
}

//GET current configuraion of the API driver
int driver_configurator::DoGetDriverConfigurationCommand(port_cfg_t *port_cfg)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "DoGetDriverConfigurationCommand";
#endif
	DBG_CFG("%s()\n", __FUNCTION__);

	port_cfg->command_code = GET_PORT_VOLATILE_CONFIG;
	return DoPortConfigurationCommand(port_cfg);
}

//SET new configuration of the API driver.
int driver_configurator::DoSetDriverConfigurationCommand(port_cfg_t *port_cfg)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "DoSetDriverConfigurationCommand";
#endif
	DBG_CFG("%s()\n", __FUNCTION__);

	port_cfg->command_code = SET_PORT_VOLATILE_CONFIG;
	return DoPortConfigurationCommand(port_cfg);
}

//Function to check correctness of 'port_cfg_t' structure.
int driver_configurator::check_port_cfg_structure(port_cfg_t *port_cfg)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "check_port_cfg_structure";
#endif
	int rc = 0;
	unsigned long tmp_active_ch = 0;
	unsigned long cumulutive_active_ch = 0;

	wandev_conf_t		*wandev_conf = &port_cfg->wandev_conf;
	sdla_fe_cfg_t		*sdla_fe_cfg = &wandev_conf->fe_cfg;
	wan_xilinx_conf_t	*wan_xilinx_conf = &wandev_conf->u.aft;
	wanif_conf_t		*wanif_conf;

	DBG_CFG("%s()\n", __FUNCTION__);

	INFO_CFG("Running Driver Configration structure check...\n");
	if(port_cfg->num_of_ifs < 1 || port_cfg->num_of_ifs > NUM_OF_E1_TIMESLOTS){
			_INFO_CFG("Invalid number of Timeslot Groups - %d! Should be between %d and %d (including).\n",
				port_cfg->num_of_ifs, 1, NUM_OF_E1_TIMESLOTS);
			return 1;
	}

	for(unsigned int i = 0; i < port_cfg->num_of_ifs; i++){
	
		//check all logic channels were configured
		wanif_conf = &port_cfg->if_cfg[i];
		if(wanif_conf->active_ch == 0){
			_INFO_CFG("Group %d - no T1/E1 channels specified! Expecting a bitmap of channels.\n", i + 1);
			rc = 1; break;
		}

#define T1_ALL_CHANNELS_BITMAP 0xFFFFFF /* channels 1-24 */

		//check the channels are valid for media type.
		if(sdla_fe_cfg->media == WAN_MEDIA_T1){
			unsigned long t1_invalid_channels = ~T1_ALL_CHANNELS_BITMAP;

			if(wanif_conf->active_ch & (t1_invalid_channels)){
				_INFO_CFG("Group %d - Invalid channels in Channel BitMap (0x%08X)!\n", i + 1,
					wanif_conf->active_ch & (t1_invalid_channels));
				rc = 2; break;
			}
		}

		//check channels are NOT in use already by another Group
		if(wanif_conf->active_ch & cumulutive_active_ch){
			_INFO_CFG("Group %d - one or more \"T1/E1 channels\" already used by another group!\n",	i + 1);
			rc = 3; break;
		}

		//update cumulative channels for the next iteration
		cumulutive_active_ch |= wanif_conf->active_ch;

		tmp_active_ch = wanif_conf->active_ch;
		if(sdla_fe_cfg->media == WAN_MEDIA_E1 && sdla_fe_cfg->frame != WAN_FR_UNFRAMED){
			//do not count the bit for channel 0! (valid only for Unframed E1)
			tmp_active_ch &= (~0x1);
		}

		if(check_mtu(wanif_conf->mtu, wanif_conf->hdlc_streaming, tmp_active_ch)){
			rc = 4; break;
		}
	}//for()

	INFO_CFG("Driver Configration structure check - %s.\n", (rc == 0 ? "OK":"FAILED"));
	return rc;
}

int driver_configurator::print_port_cfg_structure(port_cfg_t *port_cfg)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "print_port_cfg_structure";
#endif
	wandev_conf_t		*wandev_conf = &port_cfg->wandev_conf;
	sdla_fe_cfg_t		*sdla_fe_cfg = &wandev_conf->fe_cfg;
	wan_xilinx_conf_t	*wan_xilinx_conf = &wandev_conf->u.aft;
	wanif_conf_t		*wanif_conf = &port_cfg->if_cfg[0];
	
	DBG_CFG("%s()\n", __FUNCTION__);

	_INFO_CFG("\n================================================\n");

	INFO_CFG("Card Type\t: %s(%d)\n", SDLA_DECODE_CARDTYPE(wandev_conf->card_type),
		wandev_conf->card_type);/* Sangoma Card type - S514, S518 or AFT.*/

	INFO_CFG("Number of TimeSlot Groups: %d\n", port_cfg->num_of_ifs);

	INFO_CFG("MTU\t\t: %d\n",	wandev_conf->mtu);

	print_sdla_fe_cfg_t_structure(sdla_fe_cfg);

	for(unsigned int i = 0; i < port_cfg->num_of_ifs; i++){
		_INFO_CFG("\n************************************************\n");
		_INFO_CFG("Configration of Group Number %d:\n", i);
		print_wanif_conf_t_structure(&wanif_conf[i]);
	}
	return 0;
}

int driver_configurator::print_sdla_fe_cfg_t_structure(sdla_fe_cfg_t *sdla_fe_cfg)
{
	_INFO_CFG("\n################################################\n");
	INFO_CFG("MEDIA\t\t: %s\n", MEDIA_DECODE(sdla_fe_cfg));
	if(FE_MEDIA(sdla_fe_cfg) == WAN_MEDIA_T1 || FE_MEDIA(sdla_fe_cfg) == WAN_MEDIA_E1){
		INFO_CFG("Line CODE\t: %s\n", LCODE_DECODE(sdla_fe_cfg));
		INFO_CFG("Framing\t\t: %s\n", FRAME_DECODE(sdla_fe_cfg));
		INFO_CFG("Clock Mode\t: %s\n", TECLK_DECODE(sdla_fe_cfg));
		INFO_CFG("Clock Reference port: %d (0 - not used)\n", FE_REFCLK(sdla_fe_cfg));
		INFO_CFG("Signalling Insertion Mode: %s\n", TE1SIG_DECODE(sdla_fe_cfg));

		INFO_CFG("High Impedance Mode: %s\n", 
			(FE_HIMPEDANCE_MODE(sdla_fe_cfg) == WANOPT_YES ? "Yes":"No"));
		INFO_CFG("FE_RX_SLEVEL: %d\n", FE_RX_SLEVEL(sdla_fe_cfg));
	}//if()
	INFO_CFG("TDMV LAW\t: %s\n", (FE_TDMV_LAW(sdla_fe_cfg) == WAN_TDMV_MULAW ?"MuLaw":"ALaw"));

	switch(FE_MEDIA(sdla_fe_cfg))
	{
	case WAN_MEDIA_T1:
		INFO_CFG("LBO\t\t: %s\n", LBO_DECODE(sdla_fe_cfg));
		break;
	case WAN_MEDIA_E1:
		break;
	default:
		break;
	}
	return 0;
}

int driver_configurator::print_wanif_conf_t_structure(wanif_conf_t *wanif_conf)
{
	INFO_CFG("Operation Mode\t: %s\n", wanif_conf->usedby);
	INFO_CFG("Timeslot BitMap\t: 0x%08X\n", wanif_conf->active_ch);
	INFO_CFG("Line Mode\t: %s\n", 
				(wanif_conf->hdlc_streaming == WANOPT_YES ? "HDLC":"BitStream"));
	INFO_CFG("MTU\\MRU\t\t: %d\n", wanif_conf->mtu);

	return 0;
}

int driver_configurator::DriverManagementIoctl(port_management_struct_t *port_management)
{
	DWORD ln;
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "DriverManagementIoctl";
#endif

	if(DeviceIoControl(
			wp_handle,
			IoctlPortManagementCommand,
			(LPVOID)port_management,
			sizeof(port_management_struct_t),
			(LPVOID)port_management,
			sizeof(port_management_struct_t),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		ERR_CFG("%s():Error: IoctlDriverManagementCommand failed!!\n", __FUNCTION__);
		return 1;
	}else{
		return 0;
	}
}

int driver_configurator::get_hardware_info(hardware_info_t *hardware_info)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "get_hardware_info";
#endif
	port_management_struct_t port_management;

	memset(&port_management, 0x00, sizeof(port_management_struct_t));

	port_management.command_code = GET_HARDWARE_INFO;
	if(DriverManagementIoctl(&port_management)){
		return 1;
	}

	memcpy(hardware_info, port_management.data, sizeof(hardware_info_t));

	DBG_CFG("%s()\n", __FUNCTION__);

	return 0;
}

int driver_configurator::stop_or_start_port(int command)
{
#if !defined(__FUNCTION__)
const char *__FUNCTION__ = "stop_port";
#endif
	port_management_struct_t port_management;

	memset(&port_management, 0x00, sizeof(port_management_struct_t));

	port_management.command_code = command;

	if(DriverManagementIoctl(&port_management)){
		return 1;
	}

	switch(port_management.operation_status)
	{
	case SANG_STATUS_SUCCESS:
	case SANG_STATUS_CAN_NOT_STOP_DEVICE_WHEN_ALREADY_STOPPED:
		//OK
		break;
	default:
		ERR_CFG("%s: %s(): return code: %s\n", wanpipe_name_str,
			__FUNCTION__, SDLA_DECODE_SANG_STATUS(port_management.operation_status));
		break;
	}

	return port_management.operation_status;
}

int driver_configurator::push_a_card_into_wanpipe_info_array(
								wanpipe_instance_info_t *wanpipe_info_array,
								hardware_info_t *new_hw_info)
{
	unsigned int card_ind;
	hardware_info_t *tmp_hw_info;

	//check this card is not in the array already
	for(card_ind = 0; card_ind < MAX_CARDS; card_ind++){
		tmp_hw_info = &wanpipe_info_array[card_ind].hardware_info;

		if(	(new_hw_info->pci_bus_number == tmp_hw_info->pci_bus_number) &&
			(new_hw_info->pci_slot_number == tmp_hw_info->pci_slot_number)){
			//already in array. not an error.
			DBG_CFG("already in array.\n");
			return 1;
		}
	}

	//find an empty slot in the array and place the new card there
	for(card_ind = 0; card_ind < MAX_CARDS; card_ind++){
		tmp_hw_info = &wanpipe_info_array[card_ind].hardware_info;
		if(tmp_hw_info->card_model == -1){
			//found an empty slot in array
			memcpy(tmp_hw_info, new_hw_info, sizeof(hardware_info_t));
			return 0;
		}
	}

	//too many cards in the computer.
	INFO_CFG("Warning: number of cards in the computer is greater than the maximum of %d.\n", 
		MAX_CARDS);
	return 2;
}

//
//Scan all running wanpipes, check which ones are on the same card,
//figure out how many physical cards installed.
//
//Returns:	-1 - if error.
//			number of cards, starting from zero.
int driver_configurator::scan_for_sangoma_cards(wanpipe_instance_info_t *wanpipe_info_array, int card_model)
{
	int card_counter = 0, wp_ind;
	hardware_info_t	hardware_info;

	//initialize hardware information array.
	memset(wanpipe_info_array, -1, sizeof(wanpipe_instance_info_t)*MAX_CARDS);

	DBG_CFG("%s(): wanpipe_info_array: 0x%p\n", __FUNCTION__, wanpipe_info_array);

	for(wp_ind = 0; wp_ind < (MAX_CARDS	* MAX_PORTS_PER_CARD); wp_ind++){
		if(init(wp_ind, 0)){
			continue;
		}
		if(get_hardware_info(&hardware_info)){
			card_counter = -1;
			break;
		}

		DBG_CFG("card_model		: %s (0x%08X)\n",
			SDLA_ADPTR_NAME(hardware_info.card_model), hardware_info.card_model);
		DBG_CFG("firmware_version\t: 0x%02X\n", hardware_info.firmware_version);
		DBG_CFG("pci_bus_number\t\t: %d\n", hardware_info.pci_bus_number);
		DBG_CFG("pci_slot_number\t\t: %d\n", hardware_info.pci_slot_number);
		
		if(card_model == hardware_info.card_model){
			if(push_a_card_into_wanpipe_info_array(wanpipe_info_array, &hardware_info) == 0){
				//At this point we know: 1. WANPIPE number, for example WANPIPE1.
				//						 2. Card type, for example A104.
				//It allows to figure out Device Name of each port on the card.
				//For example: for A104 it will be WANPIPE1, WANPIPE2, WANPIPE3, WANPIPE4
				wanpipe_info_array[card_counter].wanpipe_number = wp_ind;
				DBG_CFG("%s(): wp_ind: %u\n", __FUNCTION__, wp_ind);
				card_counter++;
			}
		}else{
			DBG_CFG("Found a Card but Card Model is different from what was searched for.\n");
		}
	}//for()

	return card_counter;
}
