
// g3ti_api.cpp : Defines the entry point for the console application.
//

#include <iostream>
#include "SangomaCard.h"

using namespace Sangoma;

#define MAX_HANDLES	256
HANDLE port_event_handles[MAX_HANDLES];

const int VALID_NUM_OF_ARGS = 2;			///< The valid number of arguements expected by the program
const std::string A104_MODEL = "A104";		///< Command line string used to indicate to check the system for Sangoma A104 cards
const std::string A108_MODEL = "A108";		///< Command line string used to indicate to check the system for Sangoma A108 cards
bool gRunning = false;	///< Flag which is true if the sample program is currently monitoring the cards, false otherwise
	
std::vector<SangomaCard*> sangoma_cards;			//stores all the cards in the system
std::vector<SangomaPort*> sangoma_ports;			//stores all the ports in the system
std::vector<unsigned int> next_framing_type_index;	//for each port store the index of the next framing type to open the port with

#if !defined(__FUNCTION__)
#define	__FUNCTION__ "function"
#endif

#define INFO_MAIN	if(1)printf
#define DBG_MAIN	if(0)printf
#define ERR_MAIN	if(1)printf("Error: %s():", __FUNCTION__);if(1)printf

static char rx2tx = 0;

#define BUFFER_SIZE	1024*9

//SETUP FOR DETERMING PORT CONFIGURATION
const unsigned int NUMBER_OF_E1_FRAME_TYPES = 6;
const unsigned int NUMBER_OF_T1_FRAME_TYPES = 2;
const FramingType E1_TYPES[NUMBER_OF_E1_FRAME_TYPES] = { E1_CAS_CRC4_ON, E1_CAS_CRC4_OFF, E1_CCS_CRC4_ON, E1_CCS_CRC4_OFF, E1_CAS_UNFRAMED, E1_CCS_UNFRAMED };
const FramingType T1_TYPES[NUMBER_OF_T1_FRAME_TYPES] = { T1_EXTENDED_SUPERFRAME, T1_SUPERFRAME };	

char* wp_get_time_stamp()
{
    static char formatted_time_stamp[400];
    SYSTEMTIME timestamp;
    GetLocalTime(&timestamp);
	_snprintf(formatted_time_stamp, sizeof(formatted_time_stamp), "Y:%d/M:%d/D:%d. H:%d/M:%d/S:%d/MS:%u",
		timestamp.wYear, 
		timestamp.wMonth, 
		timestamp.wDay,
		timestamp.wHour, 
		timestamp.wMinute,
		timestamp.wSecond,
		timestamp.wMilliseconds);
	return formatted_time_stamp;
}   

void print_rx_data(unsigned char *buffer, int buffer_length)
{
	printf("Data length: %u\n", buffer_length);

	for(int ln = 0; ln < buffer_length; ln++){
		if((ln % 20 == 0)){
			if(ln){
				printf("\n");
			}
			printf("%04d ", ln/20);//print offset
		}
		printf("%02X ", buffer[ln]);
	}
	printf("\n");
}

static int parse_command_line_args(
	int argc, char* argv[],
	CardModel *model,
	unsigned int *number_of_ports_per_card,
	Configuration & configuration,
	BufferSettings & buffer_settings)
{
	int i;
#define USAGE_STR \
"\n\
Usage: g3ti_api CardModel [LineType] [h MaxCableLossOption]\n\
\n\
Options:\n\
\tCardModel				a104 or a108\n\
\tLineType				T1 or E1\n\
\tMaxCableLossOption	430, 360, 300, 225, 180, 175, 120\n\
\ttx_tristate			this option will disable transmitter\n\
\n\
Example 1: g3ti_api a108 e1\n\
Example 2: g3ti_api a108 t1\n\
Example 3: g3ti_api a108 e1 h 320\n\
Example 4: g3ti_api a108 e1 h 320 tx_tristate\n"

	*number_of_ports_per_card = 0;

	for(i = 0; i < argc; i++){
	
		if(_strnicmp(argv[i], "?", strlen(argv[i])) == 0 || _strnicmp(argv[i], "/?",  strlen(argv[i])) == 0){
			printf(USAGE_STR);
			return 1;
		}else if(_strnicmp(argv[i], "T1",  strlen(argv[i])) == 0){
			configuration.E1OrT1 = T1;
		}else if(_strnicmp(argv[i], "E1", strlen(argv[i])) == 0){
			configuration.E1OrT1 = E1;
		}else if(_strnicmp(argv[i], "a104",  strlen(argv[i])) == 0){
			*model = A104;
			*number_of_ports_per_card = NUMBER_OF_A104_PORTS;
		}else if(_strnicmp(argv[i], "a108",  strlen(argv[i])) == 0){
			*model = A108;
			*number_of_ports_per_card = NUMBER_OF_A108_PORTS;
		}else if(_strnicmp(argv[i], "h",  strlen(argv[i])) == 0){
			if (i+1 > argc-1){
				ERR_MAIN("No Valid 'Max Cable Loss' value was provided!\n");
				printf(USAGE_STR);
				return 1;
			}
			switch(atoi(argv[i+1]))
			{
			case MCLV_43_0dB:
			case MCLV_36_0dB:
			case MCLV_30_0dB:
			case MCLV_22_5dB:
			case MCLV_18_0dB:
			case MCLV_17_5dB:
			case MCLV_12_0dB:
				configuration.HighImpedanceMode = true;
				configuration.MaxCableLoss = (MAX_CABLE_LOSS)atoi(argv[i+1]);
				break;
			default:
				configuration.HighImpedanceMode = false;
				configuration.MaxCableLoss = MCLV_12_0dB;
			}
		}else if(_strnicmp(argv[i], "tx_tristate", strlen(argv[i])) == 0){
			configuration.TxTristateMode = true;
		
		}else if(_strnicmp(argv[i], "bm",  strlen(argv[i])) == 0){
			if (i+1 > argc-1){
				ERR_MAIN("No Valid 'Buffer Multiplier' value was provided!\n");
				printf(USAGE_STR);
				return 1;
			}
			buffer_settings.BufferMultiplierFactor = atoi(argv[i+1]);
			if(	buffer_settings.BufferMultiplierFactor < MIN_BUFFER_MULTIPLIER_FACTOR ||
				buffer_settings.BufferMultiplierFactor > MAX_BUFFER_MULTIPLIER_FACTOR){
				ERR_MAIN("The 'Buffer Multiplier' value of %d is invalid! Minimum is: %d, maximum is: %d! Setting to %d.\n",
					buffer_settings.BufferMultiplierFactor, 
					MIN_BUFFER_MULTIPLIER_FACTOR, MAX_BUFFER_MULTIPLIER_FACTOR,
					MIN_BUFFER_MULTIPLIER_FACTOR);
				buffer_settings.BufferMultiplierFactor = MIN_BUFFER_MULTIPLIER_FACTOR;
			}else{
				INFO_MAIN("User's 'Buffer Multiplier' value is: %d.\n", buffer_settings.BufferMultiplierFactor);
			}
		}else if(_strnicmp(argv[i], "rx2tx", strlen(argv[i])) == 0){
			rx2tx = 1;
		}
	}//for()

	if(*number_of_ports_per_card == 0){
		printf(USAGE_STR);
		return 1;
	}

	INFO_MAIN("\nCommand Line parsing summary:\n");
	INFO_MAIN("CardModel: %s, Line Type: %s, HighImpedance: %s, MaxCableLoss: %u, TxTristateMode: %s\n\n", 
		(*model == A104 ? "A104":"A108"), (configuration.E1OrT1 == T1 ? "T1":"E1"),
		(configuration.HighImpedanceMode == true ? "On":"Off"), configuration.MaxCableLoss,
		(configuration.TxTristateMode == true ? "On":"Off"));
	INFO_MAIN("Press any key to continue.\n");
	_getch();

	return 0;
}


///////////////////////////////////////////////////////////////////////////////////////////////
///	\fn		PortEventHandler(SangomaPort* p_port, const PortEvent & event)
///	\brief	This function handles events raised by the Sangoma classes.
///	\param	p_port	Pointer to the SangomaPort object that issued this event
///	\param	event	PortEvent stucture contain the handle to the even and its status code
///	\return	true if the event was successfully handled, false if an error occured
///	\author Josh Markwordt
///	\date	10/18/2007
///////////////////////////////////////////////////////////////////////////////////////////////
bool PortEventHandler(SangomaPort* p_port, const PortEvent & event)
{
	int number_of_remaining_buffers_in_queue = 0;
	Alarms sangoma_alarms;
	//CHECK IF WE WERE PASSED A VALID PORT
	bool invalid_port = (NULL == p_port);
	if (invalid_port) {
		std::cout << "ERROR: PortEvent created on invalid port." << std::endl;
		return false;
	}
	
	if(p_port->GetIsOpen() == false){
		/* Some other port's event get signalled. Don't try to Read() on a closed port. */
		return true;
	}

	//READ FROM THE PORT THAT GENERATED THE EVENT (THIS READ NEEDS TO BE THREAD SAFE)
	//Regardless of the reason for the event always call Read().  This will either gives us the data or return
	//reason the event was triggered (e.g. PORT_IN_SYNC, PORT_OUT_OF_SYNC).
	do{
		int read_result = p_port->Read(number_of_remaining_buffers_in_queue);

		//PROCESS THE EVENT.
		//Also check if the read was successful, if not then we need to determine what the
		//problem was.
		DBG_MAIN("Port %u: ", p_port->GetPortNumber());
		switch(read_result)
		{
		case ReadDataAvailable:
			p_port->rx_discard_counter++;
			DBG_MAIN("ReadDataAvailable\n");
			//As an example of how to access Rx data, print the Rx data:
			//print_rx_data(p_port->GetRxDataBuffer(), p_port->GetRxDataLength());
			//As an example of how to transmit data, call Write() whith Rx data as input.
			if (rx2tx) {
				//To avoid blocking receiver when waiting for transmitter,
				//discard 1-st hundred rx buffers before starting rx2tx.
				if (p_port->rx_discard_counter > 100) {
					p_port->Write(p_port->GetRxDataBuffer(), p_port->GetRxDataLength());
				}
			}
			break;
		case LineConnected:
			p_port->rx_discard_counter = 0;
			DBG_MAIN("LineConnected\n");
			/// \todo Handle the line connected event
			//DAVIDR: port is open, line connected --> start handling incoming data.
			break;
		case LineDisconnected:
			DBG_MAIN("LineDisconnected\n");
			//DAVIDR: the line got disconnected. If Alarams are of any interest, read them here.
			//May put a message for the user "cable disconnected".
			//
			sangoma_alarms = p_port->GetAlarms();
			/// \todo Handle the line disconnected event
			//DAVIDR Note: Some lines go up/down frequently and closing the port and starting Line type discovery
			//may cause a lot of unwanted load on the CPU. The best thing to do is to keep the port open
			//and wait for the LineConnected event, because the line type is already known and in the
			//real world it is not changing often. If the LineConnected event not received within
			//some timeout, it may make sense to close the port and start the Line type discovery.
			break;
		case ERROR_NO_DATA_AVILABLE:
			//FIXME: This condition may occur only during Line Discovery. Since no data can be received from
			//a disconnected port, this should be treated as a Warning, not an Error. Still, should be debugged in the future.
			//For now, ignore the warning.
			std::cout << "Port: " << p_port->GetPortNumber() << "Warning: The event triggered but no data as available\n" << std::endl;
#if DBG_READ_ERR
			//exit(1);
			//DebugBreak();
#endif
			break;
		case ERROR_INVALID_DESTINATION_BUFFER:
			std::cout << "Port: " << p_port->GetPortNumber() << "Error: Unable to read because provided buffer was invalid\n" << std::endl;
			break;
		case ERROR_INVALID_DESTINATION_BUFFER_TOO_SMALL:
			std::cout << "Port: " << p_port->GetPortNumber() << "Error: Unable to read because provided buffer was too small to hold the data.\n" << std::endl;
			break;
		case ERROR_PORT_NOT_OPEN:
			std::cout << "Port: " << p_port->GetPortNumber() << "Error: The port is not Open.\n" << std::endl;
			break;
		case ERROR_DEVICE_IOCTL:
			std::cout << p_port->GetLastError();
			break;
		default:
			std::cout << "Port: " << p_port->GetPortNumber() << "Unknown Error reported by Read()\n" << std::endl;
			break;
		} // switch(read_result)
	}while( number_of_remaining_buffers_in_queue > 0 );

	return true;
}

DWORD SangomaThreadFunc(LPVOID p_param)
{ 
    DBG_MAIN( "\n%s() - start\n", __FUNCTION__); 
	//CAST TO THE OBJECT THAT WAS PASSED IN
	std::vector<SangomaPort*>* p_ports = static_cast<std::vector<SangomaPort*>*>(p_param);

	SangomaPort		*p_port;

	//DETERMINE THE EVENT HANDLES FOR ALL THE PORTS IN ALL OF THE CARDS
	//these handles will be passed to WaitForMultipleObjects so that we know when to read
	unsigned int total_number_of_ports_in_system = static_cast<unsigned int>( p_ports->size() );
	for(unsigned int port_index = 0;
		port_index < total_number_of_ports_in_system && port_index < MAX_HANDLES;
		++port_index)
	{
		p_port = p_ports->at(port_index);
		port_event_handles[port_index] = p_port->GetEvent().EventHandle;
	}

	//SETUP VARIABLES FOR PROCESSING THE EVENT
	DWORD PORT_READ_TIMEOUT = 3000;				//arbitrarily set, doesn't matter b/c we don't do anything on a time out
	BOOL wait_for_all_events = FALSE;			//we want to be notified when any one event has occurred, TRUE waits for every single event to become signaled
	DWORD FIRST_PORT_SIGNALED = WAIT_OBJECT_0;	//WAIT_OBJECT_0 is equal to 0 and represents the first object
	DWORD LAST_PORT_SIGNALED = WAIT_OBJECT_0 + total_number_of_ports_in_system - 1;	//The WAIT_OBJECT plus the index of the handle indicates which event was signaled.

	//START PROCESSING IN THE THREAD
	while(gRunning)
	{
		//WAIT FOR AN EVENT FROM THE PORT
		//We are signalled anytime there is an event.  This could be for a line connection,
		//a line disconnection, or if there is data available.
		DWORD wait_result = WaitForMultipleObjects(	total_number_of_ports_in_system, 
													port_event_handles, 
													wait_for_all_events, 
													PORT_READ_TIMEOUT);

		//check if WaitForSingleObject returned because the event was signaled.
		bool port_signaled = (FIRST_PORT_SIGNALED <= wait_result) && (wait_result <= LAST_PORT_SIGNALED);
		if( port_signaled )
		{
			// Retreive the signaled port from the card that the event to occured on
			unsigned int signaled_port_index = static_cast<unsigned int>(wait_result);

			// Retreive the signaled port from the card that the even to occured on
			p_port = p_ports->at(signaled_port_index);
			if(p_port == NULL){
				std::cout << "ERROR: invalid port index: " << signaled_port_index << std::endl;
			}else if (PortEventHandler(p_port, p_port->GetEvent()) == false) {
				// Call the event handler to take the appropriate actions
				std::cout << "ERROR: An event on port " << signaled_port_index << " could not be handled properly." << std::endl;
			}
		} // if( port_signaled )
	
	} // while(gRunning)

    DBG_MAIN( "\n%s() - end\n", __FUNCTION__); 
    return 0; 
}

const char *decode_port_sync_status(SynchronizationStatus & status)
{
	switch( status )
	{
	case PORT_IN_SYNC: 
		return "PORT_IN_SYNC"; 
	case PORT_NOT_IN_SYNC:
		return "PORT_NOT_IN_SYNC"; 
	case PORT_WAITING_FOR_SYNC:
		return "PORT_WAITING_FOR_SYNC"; 
	}
	return "Unknown sync status";
}

void print_stats(LPVOID p_param)
{
	std::vector<SangomaPort*>* p_ports = static_cast<std::vector<SangomaPort*>*>(p_param);
	//clear the screen
	system("cls");
	
	INFO_MAIN("Time: %s\n", wp_get_time_stamp());
	for(unsigned int port_index = 0; port_index < p_ports->size(); ++port_index)
	{
		SangomaPort* p_port = p_ports->at(port_index);		//make a copy of the pointer to make it easier to work with
		SynchronizationStatus status = p_port->GetSynchronizationStatus();	//retrieve the synchronization status. DAVIDR: requires context switch to kernel-mode, should not be called too often.
		Statistics stats = p_port->GetStatistics();							//Get the statistics for this port

		INFO_MAIN("Port:%02d(unmapped:%02d):\t%20s\tRx Count: %8u\tTx Count: %8u\n",
			p_port->GetPortNumber(), p_port->GetUnmappedPortNumber(),
			decode_port_sync_status(status), stats.BytesReceived, stats.BytesSent); 
	}
	std::cout << "Hit x to exit the program." << std::endl;
}

int main_program_loop(LPVOID p_param, Configuration & configuration)
{
	//CAST TO THE OBJECT THAT WAS PASSED IN
	std::vector<SangomaPort*>* p_ports = static_cast<std::vector<SangomaPort*>*>(p_param);

	//CONTINUOUSLY TRY TO DETERMINE THE PORTS CONFIGURATION
	//if the port is in-sync nothing happens but while the port is out of sync
	//we cycle through the framing types to try and find the correct one.
	while( gRunning )
	{
		//DETERMINE EACH PORTS CONFIGURATION
		for(unsigned int port_index = 0; port_index < p_ports->size(); ++port_index)
		{
//printf("in the loop!!!!!!!!!!\n");
			//RETRIEVE THE SYNCHRONIZATION STATUS
			SangomaPort* p_port = p_ports->at(port_index);		//make a copy of the pointer to make it easier to work with
			SynchronizationStatus status = p_port->GetSynchronizationStatus();	//retrieve the synchronization status. DAVIDR: requires context switch to kernel-mode, should not be called too often.

			DBG_MAIN("Port %d: %s\n", port_index, decode_port_sync_status(status)); 
			//BASED ON THE STATUS DETERMINE WHAT TO DO
			switch( status )
			{
			case PORT_IN_SYNC: 
				{
				Configuration cfg = p_port->GetConfiguration();
				//do nothing here, read will be performed in their separate threads
				INFO_MAIN("Port %d: Got into Sync. LineType:%s, Framing: %s\n", p_port->GetPortNumber(),
					(cfg.E1OrT1==E1?"E1":"T1"), DECODE_FRAMING(cfg.Framing));
				//Sleep(1000);
				}
				break;
			case PORT_NOT_IN_SYNC:
				{
				//determine the next configuration to try
				unsigned int next_frame_type = next_framing_type_index.at(port_index);						//retrieve the next frame type to try
				if(configuration.E1OrT1 == T1){
					next_framing_type_index.at(port_index) = (next_frame_type + 1) % NUMBER_OF_T1_FRAME_TYPES;	//move the stored index to the frame type to try next time
					configuration.Framing = T1_TYPES[next_frame_type];
					configuration.LineCoding = ON;
				}else{
					next_framing_type_index.at(port_index) = (next_frame_type + 1) % NUMBER_OF_E1_FRAME_TYPES;	//move the stored index to the frame type to try next time
					configuration.Framing = E1_TYPES[next_frame_type];
					configuration.LineCoding = ON;
				}
				//open the port with the new configuration
				p_port->Close();				//close the port
				p_port->Open(configuration);	//re-open the port with the new configuration
				}
				break;
			case PORT_WAITING_FOR_SYNC:
				//do nothing because we are waiting on a configuration to find sync
				break;
			default:
				//error, unknown synchronization status
				break;
			} // switch( status )
		} // for(unsigned int port_index = 0; port_index < ports.size(); ++port_index)

		Sleep(1000);	//sleep slightly between each iteration to give the CPU some relief

		//print some stats
		print_stats(&sangoma_ports);

		//PROCESS ANY EXIT COMMANDS BY THE USER
		if(_kbhit() == 0){
			continue;
		}
		char key = _getch();
		switch(tolower(key))
		{
		case 'x':
		case 'q':
			gRunning = false;
			break;
		default:
			std::cout << "Invalid command!" << std::endl;
			break;
		}
	
	} // while( gRunning )
	
	std::cout << "Exiting ... " << std::endl;
	Sleep(5000);	//Sleep 5 seconds before exiting to make sure the other thread exits

	return 0;
}

int __cdecl main(int argc, char* argv[])
{
	unsigned int number_of_cards;
	struct BufferSettings buffer_settings;
	HANDLE	hSangomaThread;
	DWORD	SangomaThreadId;

	//create the configuration
	Configuration configuration;


	std::cout << "Application compiled on " << __DATE__ " " << __TIME__ << std::endl;
	std::cout << "Press any key to continue." << std::endl;
	_getch();

	// PROCESS THE SUPPLIED INPUT
	// Parse the command line arguements to retreive the card model to check for.  At this
	// time, we only expect to use either A104 or A108s in the system, but not a mixture of the
	// two.
	CardModel model;
	unsigned int number_of_ports_per_card = 0;
	//default line type is T1
	configuration.E1OrT1 = T1;
	configuration.HighImpedanceMode = false;
	configuration.MaxCableLoss = MCLV_12_0dB;//default value for both Hi Impedance and Normal mode
	configuration.TxTristateMode = false;

	buffer_settings.BufferMultiplierFactor = MIN_BUFFER_MULTIPLIER_FACTOR;// go to higher mulitiplier only if CPU usage is high
	buffer_settings.NumberOfBuffersPerPort = MIN_NUMBER_OF_BUFFERS_PER_API_INTERFACE;

	if(parse_command_line_args(argc, argv, &model, &number_of_ports_per_card, configuration, buffer_settings)){
		return 1;
	}

	std::cout << "Search for '" << (model==A104?"A104":"A108") << "' Sangoma card." << std::endl;

	number_of_cards = GetNumberOfCardsInstalled(model);
	INFO_MAIN("Found %d %s card(s).\n", number_of_cards, (model==A104?"A104":"A108"));

	if(number_of_cards < 1){
		ERR_MAIN("No %s Sangoma cards running. Exiting...\n", (model==A104?"A104":"A108"));
		return 1;
	}

	for (unsigned int card_number = 0; card_number < number_of_cards; card_number++) 
	{
		SangomaCard* p_card = new SangomaCard(card_number, model, buffer_settings, buffer_settings);
		sangoma_cards.push_back(p_card);

		INFO_MAIN("\nCard Driver version: %i.%i.%i.%i, Firmware version: %s, PCI Slot: %i, PCI Bus: %i\n",
			p_card->GetCardDriverVersion().Major, p_card->GetCardDriverVersion().Minor,
			p_card->GetCardDriverVersion().Build, p_card->GetCardDriverVersion().Revision,
			p_card->GetFirmwareVersion().c_str(), p_card->GetCardPciBus(), p_card->GetCardPciSlot());
		
		//Add this cards ports to the vector containing all the system's ports
		unsigned int number_of_ports = p_card->GetNumberOfPorts();
		for(unsigned int port_index = 0; port_index < number_of_ports; ++port_index)
		{
			sangoma_ports.push_back( p_card->GetPort(port_index) );
			next_framing_type_index.push_back(0);	//start with the first framing type, this indexes T1/E1_TYPES to find the frame type to use
		}
	} // for (unsigned int card_number = 0; card_number < number_of_cards; card_number++)

	gRunning = true;
    hSangomaThread = CreateThread( 
        NULL,                       /* no security attributes        */ 
        0,                          /* use default stack size        */ 
        (LPTHREAD_START_ROUTINE)SangomaThreadFunc, /* thread function     */ 
        &sangoma_ports,				/* argument to thread function   */ 
        0,                          /* use default creation flags    */ 
        &SangomaThreadId			/* returns the thread identifier */ 
		);

	if(hSangomaThread == NULL){
		DBG_MAIN( "Failed to create Sangoma thread!!\n");
		return 1;
	}

	//This thread handles real-time events, give 'time critical' priority.
	if(SetThreadPriority(hSangomaThread, THREAD_PRIORITY_TIME_CRITICAL) == FALSE){
		DBG_MAIN( "Failed to set Sangoma thread priority!!\n");
		return 1;
	}

	main_program_loop(&sangoma_ports, configuration);

	//DELETE ALL THE SANGOMA CARDS
	for(unsigned int card_index = 0; card_index < sangoma_cards.size(); ++card_index)
	{
		delete sangoma_cards.at(card_index);
	}

	INFO_MAIN("\n%s: Done.\n", argv[0]);
	return 0;
}


