//////////////////////////////////////////////////////////////////////
// sangoma_interface.cpp: interface for Sangoma API driver.
//
// Author	:	David Rokhvarg	<davidr@sangoma.com>
//////////////////////////////////////////////////////////////////////

#include "sangoma_interface.h"
#include "driver_configurator.h"

#if !defined(__FUNCTION__)
#define	__FUNCTION__ "function"
#endif

#define DBG_IFACE	if(0)printf
#define INFO_IFACE	if(0)printf
#define DBG_PORT_NUM if(0)printf

#define ERR_IFACE	if(1)printf("Error: %s():", __FUNCTION__);if(1)printf

#define MAX_RETRY_COUNTER	4

#define DO_COMMAND(wan_udp)	DoManagementCommand(hGeneralCommands, &wan_udp);

sangoma_interface::sangoma_interface(int wanpipe_num, int interface_num)
{
	DBG_IFACE("%s()\n", __FUNCTION__);

	wanpipe_number = wanpipe_num;
	interface_number = interface_num;

	DBG_PORT_NUM("%s(): (1) wanpipe_number: %i, interface_number: %i\n", 
		__FUNCTION__, wanpipe_number, interface_number);

	memset(device_name, 0x00, DEV_NAME_LEN);
	memset(&drv_version, 0x00, sizeof(drv_version));

	//////////////////////////////////////////////////////////////////
	driver_configurator drv_cfg;
	hardware_info_t hardware_info;
	card_model = A104_ADPTR_4TE1;//default
	if(drv_cfg.init(wanpipe_number, 0) == 0){
		if(drv_cfg.get_hardware_info(&hardware_info) == 0){
			DBG_IFACE("card_model		: %s (0x%08X)\n",
				SDLA_ADPTR_NAME(hardware_info.card_model), hardware_info.card_model);
			DBG_IFACE("firmware_version\t: 0x%02X\n", hardware_info.firmware_version);
			DBG_IFACE("pci_bus_number\t\t: %d\n", hardware_info.pci_bus_number);
			DBG_IFACE("pci_slot_number\t\t: %d\n", hardware_info.pci_slot_number);

			card_model = hardware_info.card_model;
		}
	}

	//G3 requred port numbers on A108 to be different from what Sangoma API provides.
	//Physical Port 0: Ports 0 and 1 (not 0 and 4)
	//Physical Port 1: Ports 2 and 3 (not 1 and 5)
	//Physical Port 2: Ports 4 and 5 (not 2 and 6)
	//Physical Port 3: Ports 6 and 7 (not 3 and 7)
	if(card_model == A108_ADPTR_8TE1){
		DBG_PORT_NUM("%s(): card_model == A108_ADPTR_8TE1\n", __FUNCTION__);
		//correct the interface_number for A108 accoringly
		switch(interface_num)
		{
		case 0:
			interface_number = 0;
			break;
		case 1:
			interface_number = 4;
			break;
		case 2:
			interface_number = 1;
			break;
		case 3:
			interface_number = 5;
			break;
		case 4:
			interface_number = 2;
			break;
		case 5:
			interface_number = 6;
			break;
		case 6:
			interface_number = 3;
			break;
		case 7:
			interface_number = 7;
			break;
		}
	}

	DBG_PORT_NUM("%s(): (2) wanpipe_number: %i, MAPPED interface_number: %i\n", 
		__FUNCTION__, wanpipe_number, interface_number);

	//Form the Interface Name from Wanpipe Number and Interface Index (i.e. WANPIPE1_IF0).
	_snprintf(device_name, DEV_NAME_LEN, "\\\\.\\WANPIPE%d_IF%d", 
		wanpipe_number + 1 + interface_number, 0); 
	DBG_PORT_NUM( "Using Device Name: %s\n", device_name);

	//////////////////////////////////////////////////////////////////
	//Temporarily open A device for general commands and get driver version.
	hGeneralCommands = open_first_interface_of_a_port();

    if(hGeneralCommands == INVALID_HANDLE_VALUE){
		ERR_IFACE( "Unable to open %s for general commands!\n", device_name);
	}else{
		read_api_driver_version();
		//we don't need the handle anymore, close it.
		CloseHandle(hGeneralCommands);
	}

	//////////////////////////////////////////////////////////////////
	hGeneralCommands = INVALID_HANDLE_VALUE;
	hRx = INVALID_HANDLE_VALUE;
	hTx = INVALID_HANDLE_VALUE;

	last_hook_state_event = WAN_EVENT_RXHOOK_ON; //default state is ON Hook
	is_rbs_monitoring_enabled = 0;

	//////////////////////////////////////////////////////////////////
	//receive stuff
	rx_frames_count = 0;
	rx_bytes_count = 0;
	//for counting frames with CRC/Abort errors
	bad_rx_frames_count = 0;

	//////////////////////////////////////////////////////////////////
	//transmit stuff
	tx_bytes_count = 0;
	tx_frames_count = 0;
	tx_test_byte = 0;

	//////////////////////////////////////////////////////////////////
	//IOCTL management structures and variables
	protocol_cb_size = sizeof(wan_mgmt_t)+sizeof(wan_cmd_t)+1;
	wan_protocol = 0;
	adapter_type = 0;
}

sangoma_interface::~sangoma_interface()
{
	DBG_IFACE("%s()\n", __FUNCTION__);
}

HANDLE sangoma_interface::open_first_interface_of_a_port()
{
	HANDLE hTmp = INVALID_HANDLE_VALUE;
	char cDevName[100];

	DBG_IFACE("%s()\n", __FUNCTION__);
	
	for(int ind = 0; ind < MAX_INTERFACES_PER_PORT; ind++){
		//Form the Interface Name from Wanpipe Number and Interface Index 
		//(i.e. WANPIPE1_IF0, WANPIPE1_IF1...).
		_snprintf(cDevName, sizeof(cDevName), "\\\\.\\WANPIPE%d_IF%d", 
			wanpipe_number + 1 + interface_number, ind); 

		hTmp = open_api_device(cDevName);
		if(hTmp != INVALID_HANDLE_VALUE){
			break;
		}
	}
	return hTmp;
}

int sangoma_interface::init(HANDLE EventHandle)
{
	DBG_IFACE("%s()\n", __FUNCTION__);
	DBG_IFACE("device_name: %s, EventHandle: 0x%p\n", device_name, EventHandle);

	////////////////////////////////////////////////////////////////////////////
	//Open device for general commands
	hGeneralCommands = open_api_device(device_name);
    if (hGeneralCommands == INVALID_HANDLE_VALUE){
		ERR_IFACE( "Unable to open %s for general commands!\n", device_name);
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//open another handle strictly for reading
	hRx = open_api_device(device_name);
    if(hRx == INVALID_HANDLE_VALUE){
		ERR_IFACE( "\nUnable to open %s for reading!\n", device_name);
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//open another handle strictly for transmission
	hTx = open_api_device(device_name);
    if(hTx == INVALID_HANDLE_VALUE){
		ERR_IFACE( "\nUnable to open %s for transmission!\n", device_name);
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//get current protocol
	if(GetWANConfig() == WAN_FALSE){
		ERR_IFACE( "Failed to get current protocol!\n");
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//get Front End Type (T1/E1/Analog...)
	if (get_fe_type(&adapter_type) == WAN_FALSE){
		ERR_IFACE( "Failed to get Front End Type!\n");
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//may need interface configuration, so get it now
	if(get_interface_configuration(&wanif_conf_struct)){
		ERR_IFACE( "Failed to get Interface Configuration!\n");
		return 1;
	}

	///////////////////////////////////////////////////////////////////////////
	REGISTER_EVENT	event;

#if WANPIPE_VERSION_MINOR1 >= 29
    event.hEvent = (HANDLE	*WP_POINTER_64)EventHandle;
#else
    event.hEvent = EventHandle;
#endif

	if(!event.hEvent ) {
         ERR_IFACE("Invalid 'EventHandle'!\n");
		 return 2;
	}

	//Provide API driver the "Shared Event" object structure.
	if(DoSetSharedEventCommand(&event)){
		//system error
		ERR_IFACE("DoSetSharedEventCommand() failed!! Check messages log.\n");
		return 3;
	}

	if(event.operation_status != SANG_STATUS_SUCCESS){
		//user error
		ERR_IFACE("DoSetSharedEventCommand() failed!! err: %d(%s).\n",
			event.operation_status, SDLA_DECODE_SANG_STATUS(event.operation_status));
		 return 4;
	}

	return 0;
}

//
//cleanup() - reverse everything what was done in init().
//
//It is very important to close ALL open
//handles, because the API will continue
//receive data until the LAST handel is closed.
int sangoma_interface::cleanup()
{
	INFO_IFACE("sangoma_interface::cleanup()\n");

	if(hGeneralCommands != INVALID_HANDLE_VALUE){
		int hndls_counter = get_open_handles_counter();
		if(hndls_counter > 0){
			INFO_IFACE( "There are %d API Handles to be closed.\n", hndls_counter);
		}

		INFO_IFACE( "Closing hGeneralCommands.\n");
		CloseHandle(hGeneralCommands);
		hGeneralCommands = INVALID_HANDLE_VALUE;
	}

	if(hRx != INVALID_HANDLE_VALUE){
		INFO_IFACE( "Closing hRx.\n");
		CloseHandle(hRx);
		hRx = INVALID_HANDLE_VALUE;
	}

	if(hTx != INVALID_HANDLE_VALUE){
		INFO_IFACE( "Closing hTx.\n");
		CloseHandle(hTx);
		hTx = INVALID_HANDLE_VALUE;
	}

	return 0;
}

HANDLE sangoma_interface::open_api_device(char *api_device_name)
{
//	DBG_IFACE("%s():%s\n", __FUNCTION__, api_device_name);

	return CreateFile(	api_device_name, 
						GENERIC_READ | GENERIC_WRITE, 
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						(LPSECURITY_ATTRIBUTES)NULL, 
						OPEN_EXISTING,
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH,
						(HANDLE)NULL
						);
}

int sangoma_interface::GetWANConfig( void )
{
	int err = WAN_TRUE;

	/* Get Protocol type */ 
	wan_udp.wan_udphdr_command	= WAN_GET_PROTOCOL;
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if (wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: Command WAN_GET_PROTOCOL failed! return code: 0x%X",
			wan_udp.wan_udphdr_return_code);
		return WAN_FALSE;
	}

	wan_protocol = wan_udp.wan_udphdr_data[0];

	INFO_IFACE( "Device %s running protocol: %s\n",
		device_name, DECODE_PROT(wan_protocol));
	 
	return err;
} 

int sangoma_interface::get_fe_type(unsigned char* adapter_type)
{
	int err = WAN_TRUE;

	/* Read Adapter Type */
	wan_udp.wan_udphdr_command = WAN_GET_MEDIA_TYPE;
	wan_udp.wan_udphdr_data[0] = WAN_MEDIA_NONE;
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: Command WAN_GET_MEDIA_TYPE failed! return code: 0x%X",
			wan_udp.wan_udphdr_return_code);
		return WAN_FALSE;
	}

	*adapter_type =	get_wan_udphdr_data_byte(0);

	INFO_IFACE( "Front End Type: ");
	switch(*adapter_type)
	{
	case WAN_MEDIA_NONE:
		INFO_IFACE( "Serial");
		break;

	case WAN_MEDIA_T1:
		INFO_IFACE( "T1");
		break;

	case WAN_MEDIA_E1:
		INFO_IFACE( "E1");
		break;

	case WAN_MEDIA_56K:
		INFO_IFACE( "56K");
		break;

	case WAN_MEDIA_FXOFXS:
		INFO_IFACE( "Aanalog");
		break;

	default:
		INFO_IFACE( "Unknown");
		err = WAN_FALSE;
	}
	INFO_IFACE( "\n");

	return err;
}

//return POINTER to data at offset 'off'
unsigned char* sangoma_interface::get_wan_udphdr_data_ptr(unsigned char off) 
{
	unsigned char *p_data = (unsigned char*)&wan_udp.wan_udphdr_data[0];
	p_data += off;
	return p_data;
}

unsigned char sangoma_interface::set_wan_udphdr_data_byte(unsigned char off, unsigned char data) 
{
	unsigned char *p_data = (unsigned char*)&wan_udp.wan_udphdr_data[0];
	p_data[off] = data;
	return 0;
}

//return DATA at offset 'off'
unsigned char sangoma_interface::get_wan_udphdr_data_byte(unsigned char off) 
{
	unsigned char *p_data = (unsigned char*)&wan_udp.wan_udphdr_data[0];
	return p_data[off];
}

int sangoma_interface::read_te1_56k_stat(unsigned long *p_alarms)
{
	unsigned char* data = NULL;
	unsigned long 	alarms = 0x00;
	sdla_te_pmon_t*	pmon = NULL;

	DBG_IFACE("%s()\n", __FUNCTION__);

	switch(adapter_type)
	{
	case WAN_MEDIA_T1:
	case WAN_MEDIA_E1:
	case WAN_MEDIA_56K:
		;//do nothing
		break;

	default:
		ERR_IFACE( "Command invalid for Adapter Type %d.\n", adapter_type);
		return 1;
	}

	memset(&wan_udp, 0x00, sizeof(wan_udp));

	/* Read T1/E1/56K alarms and T1/E1 performance monitoring counters */
	wan_udp.wan_udphdr_command = WAN_FE_GET_STAT;
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if (wan_udp.wan_udphdr_return_code != 0){
		ERR_IFACE( "Failed to read T1/E1/56K statistics.\n");
		return 1;
	}

	data = get_wan_udphdr_data_ptr(0);
	
	*p_alarms = alarms = *(unsigned long*)data;
	if (adapter_type == WAN_MEDIA_T1 || adapter_type == WAN_MEDIA_E1){
		INFO_IFACE( "***** %s: %s Alarms *****\n\n",
			device_name, (adapter_type == WAN_MEDIA_T1) ? "T1" : "E1");
		INFO_IFACE( "ALOS:\t%s\t| LOS:\t%s\n", 
				WAN_TE_ALOS_ALARM(alarms), 
				WAN_TE_LOS_ALARM(alarms));
		INFO_IFACE( "RED:\t%s\t| AIS:\t%s\n", 
				WAN_TE_RED_ALARM(alarms), 
				WAN_TE_AIS_ALARM(alarms));
		if (adapter_type == WAN_MEDIA_T1){ 
			INFO_IFACE( "YEL:\t%s\t| OOF:\t%s\n", 
					WAN_TE_YEL_ALARM(alarms), 
					WAN_TE_OOF_ALARM(alarms));
		}else{
			INFO_IFACE( "OOF:\t%s\n", 
					WAN_TE_OOF_ALARM(alarms));
		}
	}else{
		INFO_IFACE( "***** %s: 56K CSU/DSU Alarms *****\n\n\n", device_name);
	 	INFO_IFACE( "In Service:\t\t%s\tData mode idle:\t\t%s\n",
			 	INS_ALARM_56K(alarms), 
			 	DMI_ALARM_56K(alarms));
	
	 	INFO_IFACE( "Zero supp. code:\t%s\tCtrl mode idle:\t\t%s\n",
			 	ZCS_ALARM_56K(alarms), 
			 	CMI_ALARM_56K(alarms));

	 	INFO_IFACE( "Out of service code:\t%s\tOut of frame code:\t%s\n",
			 	OOS_ALARM_56K(alarms), 
			 	OOF_ALARM_56K(alarms));
		
	 	INFO_IFACE( "Valid DSU NL loopback:\t%s\tUnsigned mux code:\t%s\n",
			 	DLP_ALARM_56K(alarms), 
			 	UMC_ALARM_56K(alarms));

	 	INFO_IFACE( "Rx loss of signal:\t%s\t\n",
			 	RLOS_ALARM_56K(alarms)); 
	}

	pmon = (sdla_te_pmon_t*)&data[sizeof(unsigned long)];
/*
//FIXME:
	INFO_IFACE( "\n\n***** %s: %s Performance Monitoring Counters *****\n\n",
			device_name, (adapter_type == WAN_MEDIA_T1) ? "T1" : "E1");
	INFO_IFACE( "Framing Bit Error:\t%ld\tLine Code Violation:\t%ld\n", 
			pmon->frm_bit_error,
			pmon->lcv);
	if (adapter_type == WAN_MEDIA_T1){
		INFO_IFACE( "Out of Frame Errors:\t%ld\tBit Errors:\t\t%ld\n", 
				pmon->oof_errors,
				pmon->bit_errors);
	}else{
		INFO_IFACE( "Far End Block Errors:\t%ld\tCRC Errors:\t%ld\n", 
				pmon->far_end_blk_errors,
				pmon->crc_errors);
	}
*/
	return 0;
}

char sangoma_interface::alarms_present()
{
	unsigned long alarms;

	DBG_IFACE("%s()\n", __FUNCTION__);

	if(read_te1_56k_stat(&alarms)){
		return 1;
	}

	if (adapter_type == WAN_MEDIA_T1 || adapter_type == WAN_MEDIA_E1){
		INFO_IFACE( "***** %s: %s Alarms *****\n\n",
			device_name, (adapter_type == WAN_MEDIA_T1) ? "T1" : "E1");

		if(!strcmp(WAN_TE_ALOS_ALARM(alarms), ON_STR)){
			return 1;
		}

		if(!strcmp(WAN_TE_LOS_ALARM(alarms), ON_STR)){
			return 1;
		}

		if(!strcmp(WAN_TE_RED_ALARM(alarms), ON_STR)){
			return 1;
		}
		
		if(!strcmp(WAN_TE_AIS_ALARM(alarms), ON_STR)){
			return 1;
		}

		if (adapter_type == WAN_MEDIA_T1){ 
			if(!strcmp(WAN_TE_YEL_ALARM(alarms), ON_STR)){
				return 1;
			}

			if(!strcmp(WAN_TE_OOF_ALARM(alarms), ON_STR)){
				return 1;
			}
		}else{
			if(!strcmp(WAN_TE_OOF_ALARM(alarms), ON_STR)){
				return 1;
			}
		}
	}
	return 0;
}

int sangoma_interface::loopback_command(u_int8_t type, u_int8_t mode, u_int32_t chan_map)
{
	sdla_fe_lbmode_t	*lb;
	int			err = 0, cnt = 0;

	lb = (sdla_fe_lbmode_t*)get_wan_udphdr_data_ptr(0);
	memset(lb, 0, sizeof(sdla_fe_lbmode_t));
	lb->cmd		= WAN_FE_LBMODE_CMD_SET;
	lb->type	= type;
	lb->mode	= mode;
	lb->chan_map	= chan_map;

lb_poll_again:
	wan_udp.wan_udphdr_command	= WAN_FE_LB_MODE;
	wan_udp.wan_udphdr_data_len	= sizeof(sdla_fe_lbmode_t);
	wan_udp.wan_udphdr_return_code	= 0xaa;
	
	DO_COMMAND(wan_udp);
	
	if (wan_udp.wan_udphdr_return_code){
		err = 1;
		return err;
	}
	if (lb->rc == WAN_FE_LBMODE_RC_PENDING){

		if (!cnt) printf("Please wait ..");fflush(stdout);
		if (cnt++ < 10){
			printf(".");fflush(stdout);
			Sleep(100);
			lb->cmd	= WAN_FE_LBMODE_CMD_POLL;
			lb->rc	= 0x00;
			goto lb_poll_again;
		}
		err = 2;
		goto loopback_command_exit;
	}else if (lb->rc != WAN_FE_LBMODE_RC_SUCCESS){
		err = 3;
	}
	if (cnt) printf("\n");

loopback_command_exit:
	return err;
}

void sangoma_interface::set_lb_modes(unsigned char type, unsigned char mode)
{
	DBG_IFACE("%s()\n", __FUNCTION__);
	switch(adapter_type)
	{
	case WAN_MEDIA_T1:
	case WAN_MEDIA_E1:
	case WAN_MEDIA_56K:
		;//do nothing
		break;
	default:
		ERR_IFACE( "Command invalid for Adapter Type %d.\n", adapter_type);
		return;
	}

	if(loopback_command(type, mode, ENABLE_ALL_CHANNELS)){
		ERR_IFACE("Error: Loop Back command failed!\n");
		return;
	}

	if (adapter_type == WAN_MEDIA_T1 || adapter_type == WAN_MEDIA_E1){
		INFO_IFACE("%s %s mode ... %s!\n",
				WAN_TE1_LB_ACTION_DECODE(mode),
				WAN_TE1_LB_MODE_DECODE(type),
				(!wan_udp.wan_udphdr_return_code)?"Done":"Failed");
	}else if (adapter_type == WAN_MEDIA_DS3 || adapter_type == WAN_MEDIA_E3){
		INFO_IFACE("%s %s mode ... %s!\n",
				WAN_TE3_LB_ACTION_DECODE(mode),
				WAN_TE3_LB_TYPE_DECODE(type),
				(!wan_udp.wan_udphdr_return_code)?"Done":"Failed");
	}else{
		INFO_IFACE("%s %s mode ... %s (default)!\n",
				WAN_TE1_LB_ACTION_DECODE(mode),
				WAN_TE1_LB_MODE_DECODE(type),
				(!wan_udp.wan_udphdr_return_code)?"Done":"Failed");
	}
	return;
}

void sangoma_interface::operational_stats(net_device_stats_t *stats)
{
	wan_udp.wan_udphdr_command = READ_OPERATIONAL_STATS; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command READ_OPERATIONAL_STATS failed!(rc=0x%X)\n",
			wan_udp.wan_udphdr_return_code);
	}
	memcpy(stats, get_wan_udphdr_data_ptr(0), sizeof(net_device_stats_t));

	INFO_IFACE( "**** OPERATIONAL_STATS ****\n");

	INFO_IFACE( "\ttx_dropped\t: %lu\n", stats->tx_dropped);
	INFO_IFACE( "\ttx_packets\t: %lu\n", stats->tx_packets);
	INFO_IFACE( "\ttx_bytes\t: %lu\n", stats->tx_bytes);
	INFO_IFACE( "\n");

	//receive stats
	INFO_IFACE( "\trx_packets\t: %lu\n", stats->rx_packets);
	INFO_IFACE( "\trx_bytes\t: %lu\n", stats->rx_bytes);

	INFO_IFACE( "\trx_errors\t: %lu\n", stats->rx_errors);
	INFO_IFACE( "\trx_frame_errors\t: %lu\n", stats->rx_frame_errors);

	INFO_IFACE( "\trx_packets_discarded_rx_q_full\t: %lu\n",
		stats->rx_dropped);
	INFO_IFACE( "\n");

	//hardware level stats
	//AFT card
	INFO_IFACE( "\trx_fifo_errors\t: %lu\n", stats->rx_fifo_errors);
	INFO_IFACE( "\ttx_fifo_errors\t: %lu\n", stats->tx_fifo_errors);

	//BitStream mode only - counter of data blocks transmitted automatically,
	//because no User Data was available. Transmit underrun counter from user
	//point of view.
	INFO_IFACE( "\ttx_idle_data\t: %lu\n", stats->tx_idle_packets);
}

void sangoma_interface::flush_operational_stats (void)
{
	wan_udp.wan_udphdr_command = FLUSH_OPERATIONAL_STATS; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command FLUSH_OPERATIONAL_STATS failed!\n");
	}

	INFO_IFACE( "Command FLUSH_OPERATIONAL_STATS was successful.\n");
}

void sangoma_interface::flush_tx_buffers (void)
{
	wan_udp.wan_udphdr_command = WANPIPEMON_FLUSH_TX_BUFFERS; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command WANPIPEMON_FLUSH_TX_BUFFERS failed!\n");
	}

	INFO_IFACE( "Command WANPIPEMON_FLUSH_TX_BUFFERS was successful.\n");
}

void sangoma_interface::set_idle_tx_data_buffer(unsigned char idle_tx_test_byte)
{
	//set data in Idle Tx buffer in the driver
	tx_data.api_header.data_length = 512;//should be the same as MTU/MRU
	tx_data.api_header.operation_status = SANG_STATUS_TX_TIMEOUT;
	//set the actual data for transmission
	memset(tx_data.data, idle_tx_test_byte, tx_data.api_header.data_length);

	if(DoSetIdleTxBufferCommand(hGeneralCommands, &tx_data)){
		//error
		ERR_IFACE("DoSetIdleTxBufferCommand() failed!! Check messages log.\n");
		return;
	}

	if(tx_data.api_header.operation_status != SANG_STATUS_SUCCESS){
		//error
		ERR_IFACE("DoSetIdleTxBufferCommand() failed!!.\n");
		return;
	}
}

//set RBS to be transmitted
char sangoma_interface::set_rbs(rbs_management_t *rbs_management_ptr)
{
	wan_udp.wan_udphdr_command = WAN_SET_RBS_BITS;
	memcpy(&wan_udp.wan_udphdr_data[0], rbs_management_ptr, sizeof(rbs_management_t));
	wan_udp.wan_udphdr_data_len = sizeof(rbs_management_t);

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command WAN_SET_RBS_BITS failed!\n");
		return 1;
	}

	INFO_IFACE( "**** WAN_SET_RBS_BITS OK ****\n");

	//INFO_IFACE( "ABCD_bits (HEX): 0x%02X\n", rbs_management_ptr->ABCD_bits);

	INFO_IFACE( "Channel: %d, TX RBS: A:%1d B:%1d C:%1d D:%1d\n",
			rbs_management_ptr->channel, 
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_A) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_B) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_C) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_D) ? 1 : 0);
	return 0;
}

int sangoma_interface::get_interface_configuration(if_cfg_t *wanif_conf_ptr)
{
	wan_udp.wan_udphdr_command = READ_CONFIGURATION;
	wan_udp.wan_udphdr_data_len = sizeof(if_cfg_t);

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command READ_CONFIGURATION failed!\n");
		return 1;
	}
	memcpy(wanif_conf_ptr, get_wan_udphdr_data_ptr(0), sizeof(if_cfg_t));

	INFO_IFACE( "**** READ_CONFIGURATION ****\n");
	INFO_IFACE( "Operational Mode\t: %s\n", wanif_conf_ptr->usedby);
	INFO_IFACE( "Active Channels \t: 0x%08X\n", wanif_conf_ptr->active_ch);
	INFO_IFACE( "Interface Number\t: %u\n", wanif_conf_ptr->interface_number);
	INFO_IFACE( "Media type\t\t: %u\n", wanif_conf_ptr->media);
	if(wanif_conf_ptr->media == WAN_MEDIA_FXOFXS){
		switch(wanif_conf_ptr->sub_media)
		{
		case MOD_TYPE_FXS:
			INFO_IFACE( "Media sub-type\t\t: FXS\n");
			break;
		case MOD_TYPE_FXO:
			INFO_IFACE( "Media sub-type\t\t: FXO\n");
			break;
		default:
			INFO_IFACE( "Media sub-type\t\t: Unknown!!\n");
			break;
		}
	}
	INFO_IFACE( "****************************\n");

	return 0;
}

int sangoma_interface::get_api_driver_version (PDRIVER_VERSION p_drv_version)
{
	memcpy(p_drv_version, &drv_version, sizeof(DRIVER_VERSION));

	INFO_IFACE("\nAPI version\t: %d,%d,%d,%d\n",
		p_drv_version->major, p_drv_version->minor, p_drv_version->minor1, p_drv_version->minor2);
	INFO_IFACE("\n");
	return 0;
}

int sangoma_interface::read_api_driver_version()
{
	PDRIVER_VERSION version;
	wan_udp.wan_udphdr_command = READ_CODE_VERSION; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE("Error: command READ_CODE_VERSION failed!\n");
		return 1;
	}

	version = (PDRIVER_VERSION)get_wan_udphdr_data_ptr(0);
	memcpy(&drv_version, version, sizeof(DRIVER_VERSION));

	INFO_IFACE("\nAPI version\t: %d,%d,%d,%d\n",
		version->major, version->minor, version->minor1, version->minor2);
	INFO_IFACE("\n");

	INFO_IFACE("Command READ_CODE_VERSION was successful.\n");

	// verify that Driver Version is the same as Header files used to compile this application
	if(drv_version.major != WANPIPE_VERSION_MAJOR){
		ERR_IFACE("Error: drv_version.major (%d) != WANPIPE_VERSION_MAJOR (%d)\n", drv_version.major, WANPIPE_VERSION_MAJOR);
		exit(1);
	}

	if(drv_version.minor != WANPIPE_VERSION_MINOR){
		ERR_IFACE("Error: drv_version.minor (%d) != WANPIPE_VERSION_MINOR (%d)\n", drv_version.minor, WANPIPE_VERSION_MINOR);
		exit(1);
	}

	if(drv_version.minor1 != WANPIPE_VERSION_MINOR1){
		ERR_IFACE("Error: drv_version.minor1 (%d) != WANPIPE_VERSION_MINOR1 (%d)\n", drv_version.minor1, WANPIPE_VERSION_MINOR1);
		exit(1);
	}

	if(drv_version.minor2 != WANPIPE_VERSION_MINOR2){
		ERR_IFACE("Error: drv_version.minor2 (%d) != WANPIPE_VERSION_MINOR2 (%d)\n", drv_version.minor2, WANPIPE_VERSION_MINOR2);
		exit(1);
	}

	return 0;
}

int sangoma_interface::get_open_handles_counter()
{
	wan_udp.wan_udphdr_command = GET_OPEN_HANDLES_COUNTER; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		ERR_IFACE( "Error: command GET_OPEN_HANDLES_COUNTER failed!\n");
		return -1;
	}

	return *(int*)&wan_udp.wan_udphdr_data[0];
}

// Blocking write command.
UCHAR sangoma_interface::DoWriteCommand(HANDLE drv, TX_DATA_STRUCT * pTx, UINT buffer_length)
{
	DWORD ln;

	if(DeviceIoControl(
			drv,
			IoctlWriteCommand,
			(LPVOID)pTx,
			(ULONG)buffer_length,
			(LPVOID)pTx,
			(ULONG)buffer_length,
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		return 1;
	}else{
		return 0;
	}
}

// Blocking read command.
USHORT sangoma_interface::DoReadCommand(HANDLE drv, RX_DATA_STRUCT * pRx, UINT buffer_length)
{
	DWORD ln;

	if (DeviceIoControl(
			drv,
			IoctlReadCommand,
			(LPVOID)NULL,
			0L,
			(LPVOID)pRx,
			buffer_length,
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		return 1;
	}else{
		return 0;
	}
}

// set data in Idle Transmit buffer (BitStream only).
UCHAR sangoma_interface::DoSetIdleTxBufferCommand(HANDLE drv, TX_DATA_STRUCT * pTx)
{
	DWORD ln;

	if(DeviceIoControl(
			drv,
			IoctlSetIdleTxBuffer,
			(LPVOID)pTx,
			(ULONG)sizeof(TX_DATA_STRUCT),
			(LPVOID)pTx,
			sizeof(TX_DATA_STRUCT),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		return 1;
	}else{
		return 0;
	}
}

int sangoma_interface::DoManagementCommand(HANDLE drv, wan_udp_hdr_t* wan_udp)
{
	DWORD ln;
    static unsigned char id = 0;

	wan_udp->wan_udphdr_request_reply = 0x01;
	wan_udp->wan_udphdr_id = id;
   	wan_udp->wan_udphdr_return_code = WAN_UDP_TIMEOUT_CMD;

	if(DeviceIoControl(
			drv,
			IoctlManagementCommand,
			(LPVOID)wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPVOID)wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		return 1;
	}else{
		return 0;
	}
}

UCHAR sangoma_interface::DoSetSharedEventCommand(PREGISTER_EVENT event)
{
	DWORD ln;
	
	if(DeviceIoControl(
		hRx,
		IoctlSetSharedEvent,
		(LPVOID)event,
		sizeof(REGISTER_EVENT),
		(LPVOID)event,
		sizeof(REGISTER_EVENT),
		(LPDWORD)(&ln),
		(LPOVERLAPPED)NULL
					) == FALSE){
		//check messages log
		return 1;
	}else{
		return 0;
	}
}



