//////////////////////////////////////////////////////////////////////
// sangoma_interface.h: interface for Sangoma API driver.
//
// Author	:	David Rokhvarg	<davidr@sangoma.com>
//////////////////////////////////////////////////////////////////////

#ifndef SANGOMA_INTERFACE_H
#define SANGOMA_INTERFACE_H

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <stddef.h>		//for offsetof()
#include <math.h>

#define __WINDOWS__
#include <wanpipe_defines.h>	//for 'wan_udp_hdr_t'
#include <wanpipe_version.h>	//for DRIVER_VERSION structure
#include <sang_status_defines.h>//return codes
#include <sang_api.h>			//for IOCTL codes
#include <sdla_remora.h>		//Analog card
#include <wanpipe_cfg.h>		//for RBS commands, Media Types (WAN_MEDIA_NONE...)
#include <wanpipe_tdm_api_iface.h>	//for TDMV API
#include <wanpipe_api.h>		//for Wanpipe API

#define DBG_READ_ERR 0

/**
  *@author David Rokhvarg
  */
struct func_protocol {
	int		protocol_id;
	unsigned char   prot_name[10];
	unsigned char	mbox_offset;
};


static struct func_protocol function_lookup[] = {
	{ WANCONFIG_FR,			"fr",		0 }, 
	{ WANCONFIG_MFR,		"fr",		0 },
	{ WANCONFIG_PPP,		"ppp",		0 }, 
	{ WANCONFIG_CHDLC,		"chdlc",	2 }, 
	{ WANCONFIG_ATM,		"atm",		2 }, 
	{ WANCONFIG_AFT,		"aft",		0 },//supported
	{ WANCONFIG_AFT_TE1,	"aft",		0 },//supported
	{ WANCONFIG_AFT_ANALOG,	"aft",		0 },//supported
	{ WANCONFIG_ADSL,		"adsl",		2 }, 
	{ WANCONFIG_SS7,		"ss7",		2 },
	{ WANCONFIG_X25, 		"x25",		0 }, 
	{ WANCONFIG_BITSTRM,	"bitstrm",	0 },
	{ WANCONFIG_BSC, 		"bsc",		0 }, 
	{ WANCONFIG_HDLC, 		"hdlc",		0 }, 
	{ 0,					"N/A",		0 }, 
};

#define DECODE_PROT(prot)((prot==WANCONFIG_FR)?"Frame Relay":	\
			  (prot==WANCONFIG_MFR)?"Frame Relay":				\
	          (prot==WANCONFIG_PPP)?"PPP Point-to-Point":		\
	          (prot==WANCONFIG_CHDLC)?"Cisco HDLC":				\
	          (prot==WANCONFIG_X25)?"X25":						\
	          (prot==WANCONFIG_ADSL)?"ADSL: Eth,IP,PPP/ATM":	\
			  (prot==WANCONFIG_ATM)?"ATM":						\
			  (prot==WANCONFIG_AFT)?"AFT T1/E1":				\
			  (prot==WANCONFIG_AFT_TE1)?"AFT T1/E1":			\
			  (prot==WANCONFIG_AFT_ANALOG)?"AFT Analog":		\
							"Unknown")

#define DEV_NAME_LEN			100

#define ON_STR "ON"
#define OFF_STR "OFF"

#define MAX_INTERFACES_PER_PORT 32

class sangoma_interface {

	//////////////////////////////////////////////////////////////////
	HANDLE		hGeneralCommands;
	HANDLE		hRx;
	HANDLE		hTx;

	//////////////////////////////////////////////////////////////////
	//receive stuff
	ULONG		rx_frames_count;
	ULONG		rx_bytes_count;
	//for counting frames with CRC/Abort errors
	ULONG		bad_rx_frames_count;

	//////////////////////////////////////////////////////////////////
	//transmit stuff
	ULONG		tx_bytes_count;
	ULONG		tx_frames_count;
	UCHAR		tx_test_byte;

	//////////////////////////////////////////////////////////////////
	//IOCTL management structures and variables
	wan_udp_hdr_t	wan_udp;
	int				protocol_cb_size;
	int 			wan_protocol;
	unsigned char	adapter_type;
	unsigned int	card_model;

	int GetWANConfig( void );
	int get_interface_configuration(if_cfg_t *wanif_conf_ptr);
	//return DATA at offset 'off'
	unsigned char get_wan_udphdr_data_byte(unsigned char off);
	//return POINTER to data at offset 'off'
	unsigned char *get_wan_udphdr_data_ptr(unsigned char off);
	unsigned char set_wan_udphdr_data_byte(unsigned char off, unsigned char data);
	int get_fe_type(unsigned char* adapter_type);
	int read_api_driver_version();

	int wanpipe_api_ioctl(wan_cmd_api_t *api_cmd);

	USHORT	DoReadCommand(HANDLE drv, RX_DATA_STRUCT * pRx, UINT buffer_length);
	UCHAR	DoWriteCommand(HANDLE drv, TX_DATA_STRUCT * pTx, UINT buffer_length);
	UCHAR	DoSetIdleTxBufferCommand(HANDLE drv, TX_DATA_STRUCT *pTx);
	int		DoManagementCommand(HANDLE drv, wan_udp_hdr_t* packet);
	UCHAR	DoSetSharedEventCommand(PREGISTER_EVENT event);

	//////////////////////////////////////////////////////////////////
	DRIVER_VERSION	drv_version;
	unsigned char	last_hook_state_event;
	char			is_rbs_monitoring_enabled;

	TX_DATA_STRUCT			tx_data;
	char					device_name[DEV_NAME_LEN];	
	if_cfg_t				wanif_conf_struct;

	int wanpipe_number, interface_number;

	//In Legacy Mode (G3) 1-st interface is zero-based (i.g. WANPIPE1_IF0).
	//In Regular Mode 1-st interface is one-based (i.g. WANPIPE1_IF1).
	//The Setup installs in Regular Mode.
	HANDLE open_first_interface_of_a_port();

public:

	//////////////////////////////////////////////////////////////////
	//methods
	sangoma_interface(int wanpipe_number, int interface_number);
	~sangoma_interface();

	int init(HANDLE EventHandle);
	int cleanup();

	int read_te1_56k_stat(unsigned long *p_alarms);
	char alarms_present();
	void set_lb_modes(unsigned char type, unsigned char mode);
	int loopback_command(u_int8_t type, u_int8_t mode, u_int32_t chan_map);
	
	void operational_stats (net_device_stats_t *stats);
	void flush_operational_stats (void);
	void flush_tx_buffers (void);

	void enable_rbs_monitoring();
	char set_rbs(rbs_management_t *rbs_management_ptr);

	void set_idle_tx_data_buffer(unsigned char idle_tx_test_byte);
	int get_open_handles_counter();
	HANDLE open_api_device(char *api_device_name);

	unsigned char get_adapter_type()
	{
		return adapter_type;
	}

	unsigned int get_sub_media()
	{
		return wanif_conf_struct.sub_media;
	}

	//Returns Port Number mapped specifically for G3 use.
	unsigned int get_mapped_port_number()
	{
		return interface_number;
	}

	int get_api_driver_version (PDRIVER_VERSION p_drv_version);

	UCHAR transmit(TX_DATA_STRUCT *pTx, UINT buffer_length)
	{
		return DoWriteCommand(hTx, pTx, buffer_length);
	}

	USHORT receive(RX_DATA_STRUCT *pRx, UINT buffer_length, int& number_of_remaining_buffers_in_queue)
	{	
		USHORT return_code;
		
		number_of_remaining_buffers_in_queue = 0;

		return_code = DoReadCommand(hRx, pRx, buffer_length);

		if(!return_code){
			switch(pRx->api_header.operation_status)
			{
			case SANG_STATUS_RX_DATA_AVAILABLE:
				number_of_remaining_buffers_in_queue = pRx->api_header.rx_h.current_number_of_frames_in_rx_queue;
				break;
			}
		}
		return return_code;
	}
};

#endif//#define SANGOMA_INTERFACE_H

